#ifndef CONVERSIONS_H
#define CONVERSIONS_H

// Conversions functions from OpenCV, Linux V4L samples...

// LIMIT: convert a 16.16 fixed-point value to a byte, with clipping.
#ifndef LIMIT
#define LIMIT(x) ((x)>0xffffff?0xff: ((x)<=0xffff?0:((x)>>16)))
#endif // LIMIT

#ifndef SAT
#define SAT(c) \
	if ((c) & (~255)) { if ((c) < 0) (c) = 0; else (c) = 255; }
#endif // SAT

// Color space conversion coefficients taken from the excellent
// http://www.inforamp.net/~poynton/ColorFAQ.html
//
// To avoid floating point arithmetic, the color conversion
// coefficients are scaled into 16.16 fixed-point integers.
// They were determined as follows:
//
//  double brightness = 1.0;  (0->black; 1->full scale)
//  double saturation = 1.0;  (0->greyscale; 1->full color)
//  double fixScale = brightness * 256 * 256;
//  int rvScale = (int)(1.402 * saturation * fixScale);
//  int guScale = (int)(-0.344136 * saturation * fixScale);
//  int gvScale = (int)(-0.714136 * saturation * fixScale);
//  int buScale = (int)(1.772 * saturation * fixScale);
//  int yScale = (int)(fixScale);
static __inline__ void yuv_2_rgb_pix(int y, int u, int v, 
	unsigned char* red, unsigned char* green, unsigned char* blue)
{
	int r = 0, g = 0, b = 0, y0 = 0, u0 = 0, v0 = 0, cr = 0, cg = 0, cb = 0;

	y0 = y<<16; u0 = u-128; v0 = v-128; cr = 91881*v0; cg = 22553*u0+46801*v0; cb = 116129*u0;
	r = LIMIT(y0+cr); g = LIMIT(y0-cg); b = LIMIT(y0+cb);
	*red = (unsigned char)r; *green = (unsigned char)g; *blue = (unsigned char)b;
}

static __inline__ void yuv_2_rgb_pix0(int y, int u, int v, 
	unsigned char* red, unsigned char* green, unsigned char* blue)
{
	int r = 0, g = 0, b = 0, u0 = 0, v0 = 0, cr = 0, cg = 0, cb = 0;

	u0 = u-128; v0 = v-128; cr = (359*v0)>>8; cg = (88*u0+183*v0)>>8; cb = (454*u0)>>8;
	r = y+cr; g = y-cg; b = y+cb; SAT(r); SAT(g); SAT(b);
	//r = (int)(y+1.402*(v-128)); g = (int)(y-0.34414*(u-128)-0.71414*(v-128)); b = (int)(y+1.772*(u-128)); 
	//SAT(r); SAT(g); SAT(b);
	*red = (unsigned char)r; *green = (unsigned char)g; *blue = (unsigned char)b;
}

// Turn a YUV4:2:0 block into an RGB block
//
// Video4Linux seems to use the blue, green, red channel
// order convention-- rgb[0] is blue, rgb[1] is green, rgb[2] is red.
//
// Color space conversion coefficients taken from the excellent
// http://www.inforamp.net/~poynton/ColorFAQ.html
// In his terminology, this is a CCIR 601.1 YCbCr -> RGB.
// Y values are given for all 4 pixels, but the U (Pb)
// and V (Pr) are assumed constant over the 2x2 block.
//
// To avoid floating point arithmetic, the color conversion
// coefficients are scaled into 16.16 fixed-point integers.
// They were determined as follows:
//
//  double brightness = 1.0;  (0->black; 1->full scale)
//  double saturation = 1.0;  (0->greyscale; 1->full color)
//  double fixScale = brightness * 256 * 256;
//  int rvScale = (int)(1.402 * saturation * fixScale);
//  int guScale = (int)(-0.344136 * saturation * fixScale);
//  int gvScale = (int)(-0.714136 * saturation * fixScale);
//  int buScale = (int)(1.772 * saturation * fixScale);
//  int yScale = (int)(fixScale);
static __inline__ void move420block(int yTL, int yTR, int yBL, int yBR, 
	int u, int v, int rowPixels, unsigned char * rgb)
{
	const int rvScale = 91881;
	const int guScale = -22553;
	const int gvScale = -46801;
	const int buScale = 116129;
	const int yScale  = 65536;
	int r, g, b;

	g = guScale * u + gvScale * v;
	//  if (force_rgb) {
	//      r = buScale * u;
	//      b = rvScale * v;
	//  } else {
	r = rvScale * v;
	b = buScale * u;
	//  }

	yTL *= yScale; yTR *= yScale;
	yBL *= yScale; yBR *= yScale;

	// Write out top two pixels.
	rgb[0] = LIMIT(r+yTL); rgb[1] = LIMIT(g+yTL); rgb[2] = LIMIT(b+yTL);

	rgb[3] = LIMIT(r+yTR); rgb[4] = LIMIT(g+yTR); rgb[5] = LIMIT(b+yTR);

	// Skip down to next line to write out bottom two pixels.
	rgb += 3 * rowPixels;
	rgb[0] = LIMIT(r+yBL); rgb[1] = LIMIT(g+yBL); rgb[2] = LIMIT(b+yBL);

	rgb[3] = LIMIT(r+yBR); rgb[4] = LIMIT(g+yBR); rgb[5] = LIMIT(b+yBR);
}

static __inline__ void move411block(int yTL, int yTR, int yBL, int yBR, 
	int u, int v, unsigned char * rgb)
{
	const int rvScale = 91881;
	const int guScale = -22553;
	const int gvScale = -46801;
	const int buScale = 116129;
	const int yScale  = 65536;
	int r, g, b;

	g = guScale * u + gvScale * v;
	//  if (force_rgb) {
	//      r = buScale * u;
	//      b = rvScale * v;
	//  } else {
	r = rvScale * v;
	b = buScale * u;
	//  }

	yTL *= yScale; yTR *= yScale;
	yBL *= yScale; yBR *= yScale;

	// Write out top two first pixels.
	rgb[0] = LIMIT(r+yTL); rgb[1] = LIMIT(g+yTL); rgb[2] = LIMIT(b+yTL);

	rgb[3] = LIMIT(r+yTR); rgb[4] = LIMIT(g+yTR); rgb[5] = LIMIT(b+yTR);

	// Write out top two last pixels.
	rgb += 6;
	rgb[0] = LIMIT(r+yBL); rgb[1] = LIMIT(g+yBL); rgb[2] = LIMIT(b+yBL);

	rgb[3] = LIMIT(r+yBR); rgb[4] = LIMIT(g+yBR); rgb[5] = LIMIT(b+yBR);
}

// FOURCC ?.
// Consider a YUV420 image of 6x2 pixels.
//
// A B C D U1 U2
// I J K L V1 V2
//
// The U1/V1 samples correspond to the ABIJ pixels.
//     U2/V2 samples correspond to the CDKL pixels.
//
// Convert from interlaced YUV420 to RGB24.
// [FD] untested...
static __inline__ void yuv420_2_rgb24(int width, int height, unsigned char *pIn0, unsigned char *pOut0)
{
	const int bytes = 24 >> 3;
	int i, j, y00, y01, y10, y11, u, v;
	unsigned char *pY = pIn0;
	unsigned char *pU = pY + 4;
	unsigned char *pV = pU + width;
	unsigned char *pOut = pOut0;

	for (j = 0; j <= height - 2; j += 2) {
		for (i = 0; i <= width - 4; i += 4) {
			y00 = *pY;
			y01 = *(pY + 1);
			y10 = *(pY + width);
			y11 = *(pY + width + 1);
			u = (*pU++) - 128;
			v = (*pV++) - 128;

			move420block(y00, y01, y10, y11, u, v, width, pOut);

			pY += 2;
			pOut += 2 * bytes;

			y00 = *pY;
			y01 = *(pY + 1);
			y10 = *(pY + width);
			y11 = *(pY + width + 1);
			u = (*pU++) - 128;
			v = (*pV++) - 128;

			move420block(y00, y01, y10, y11, u, v, width, pOut);

			pY += 4; // skip UV
			pOut += 2 * bytes;

		}
		pY += width;
		pOut += width * bytes;
	}
}

// FOURCCs "YU12", "I420", "IYUV".
// Consider a YUV420P image of 8x2 pixels.
//
// A plane of Y values    A B C D E F G H
//                        I J K L M N O P
//
// A plane of U values    1   2   3   4
// A plane of V values    1   2   3   4 ....
//
// The U1/V1 samples correspond to the ABIJ pixels.
//     U2/V2 samples correspond to the CDKL pixels.
//
// Convert from planar YUV420P to RGB24.
static __inline__ void yuv420p_2_rgb24(int width, int height, unsigned char *pIn0, unsigned char *pOut0)
{
	const int numpix = width * height;
	const int bytes = 24 >> 3;
	int i, j, y00, y01, y10, y11, u, v;
	unsigned char *pY = pIn0;
	unsigned char *pU = pY + numpix;
	unsigned char *pV = pU + numpix / 4;
	unsigned char *pOut = pOut0;

	for (j = 0; j <= height - 2; j += 2) {
		for (i = 0; i <= width - 2; i += 2) {
			y00 = *pY;
			y01 = *(pY + 1);
			y10 = *(pY + width);
			y11 = *(pY + width + 1);
			u = (*pU++) - 128;
			v = (*pV++) - 128;

			move420block(y00, y01, y10, y11, u, v, width, pOut);

			pY += 2;
			pOut += 2 * bytes;

		}
		pY += width;
		pOut += width * bytes;
	}
}

// FOURCC "YV12".
// To test and optimize...
static __inline__ void yvu420p_2_rgb24(int width, int height, unsigned char *pIn0, unsigned char *pOut0)
{
	const int numpix = width * height;
	const int bytes = 24 >> 3;
	int i, j, y00, y01, y10, y11, u, v;
	unsigned char *pY = pIn0;
	unsigned char *pV = pY + numpix;
	unsigned char *pU = pV + numpix / 4;
	unsigned char *pOut = pOut0;

	for (j = 0; j <= height - 2; j += 2) {
		for (i = 0; i <= width - 2; i += 2) {
			y00 = *pY;
			y01 = *(pY + 1);
			y10 = *(pY + width);
			y11 = *(pY + width + 1);
			v = (*pV++) - 128;
			u = (*pU++) - 128;

			move420block(y00, y01, y10, y11, u, v, width, pOut);

			pY += 2;
			pOut += 2 * bytes;

		}
		pY += width;
		pOut += width * bytes;
	}
}

// FOURCC "411P".
// Consider a YUV411P image of 8x2 pixels.
//
// A plane of Y values    A B C D E F G H
//                        I J K L M N O P
//
// A plane of U values    1       2
//                        3       4
//
// A plane of V values    1       2
//                        3       4
//
// The U1/V1 samples correspond to the ABCD pixels.
//     U2/V2 samples correspond to the EFGH pixels.
//
// Convert from planar YUV411P to RGB24.
// [FD] untested... Does not seem to be OK or maybe bad driver...
static __inline__ void yuv411p_2_rgb24(int width, int height, unsigned char *pIn0, unsigned char *pOut0)
{
	const int numpix = width * height;
	const int bytes = 24 >> 3;
	int i, j, y00, y01, y10, y11, u, v;
	unsigned char *pY = pIn0;
	unsigned char *pU = pY + numpix;
	unsigned char *pV = pU + numpix / 4;
	unsigned char *pOut = pOut0;

	for (j = 0; j <= height; j++) {
		for (i = 0; i <= width - 4; i += 4) {
			y00 = *pY;
			y01 = *(pY + 1);
			y10 = *(pY + 2);
			y11 = *(pY + 3);
			u = (*pU++) - 128;
			v = (*pV++) - 128;

			move411block(y00, y01, y10, y11, u, v, pOut);

			pY += 4;
			pOut += 4 * bytes;

		}
	}
}

// FOURCCs "YUYV", "YUY2", "V422", "YUNV".
// For tests...
static __inline__ void yuy2_2_rgb24(int width, int height, unsigned char *src, unsigned char *dst)
{
	unsigned char* buf = src;
	unsigned char* rgb = dst;
	int y1 = 0, y2 = 0, u = 0, v = 0;
	unsigned char r = 0, g = 0, b = 0;
	int k = width*height-1;

	while (k > 0)
	{
		k -= 2;

		y1 = *buf++;
		u = *buf++;
		y2 = *buf++;
		v = *buf++;

		yuv_2_rgb_pix(y1, u, v, &r, &g, &b);

		// Pixel width*height-3-k.
		*rgb++ = r;
		*rgb++ = g;
		*rgb++ = b;

		yuv_2_rgb_pix(y2, u, v, &r, &g, &b);

		// Pixel width*height-2-k.
		*rgb++ = r;
		*rgb++ = g;
		*rgb++ = b;
	}
}

// FOURCCs "YUYV", "YUY2", "V422", "YUNV".
// Convert from 4:2:2 YUYV interlaced to RGB24.
// Based on ccvt_yuyv_bgr32() from camstream.
static __inline__ void yuyv_2_rgb24(int width, int height, unsigned char *src, unsigned char *dst)
{
	unsigned char *s;
	unsigned char *d;
	int l, c;
	int r, g, b, cr, cg, cb, y1, y2;

	l = height;
	s = src;
	d = dst;
	while (l--) {
		c = width >> 1;
		while (c--) {
			y1 = *s++;
			cb = ((*s - 128) * 454) >> 8;
			cg = (*s++ - 128) * 88;
			y2 = *s++;
			cr = ((*s - 128) * 359) >> 8;
			cg = (cg + (*s++ - 128) * 183) >> 8;

			r = y1 + cr;
			g = y1 - cg;
			b = y1 + cb;
			SAT(r);
			SAT(g);
			SAT(b);

			*d++ = r;
			*d++ = g;
			*d++ = b;

			r = y2 + cr;
			g = y2 - cg;
			b = y2 + cb;
			SAT(r);
			SAT(g);
			SAT(b);

			*d++ = r;
			*d++ = g;
			*d++ = b;
		}
	}
}

// FOURCCs "UYVY", "Y422", "UYNV", "HDYC".
static __inline__ void uyvy_2_rgb24(int width, int height, unsigned char *src, unsigned char *dst)
{
	unsigned char *s;
	unsigned char *d;
	int l, c;
	int r, g, b, cr, cg, cb, y1, y2;

	l = height;
	s = src;
	d = dst;
	while (l--) {
		c = width >> 1;
		while (c--) {
			cb = ((*s - 128) * 454) >> 8;
			cg = (*s++ - 128) * 88;
			y1 = *s++;
			cr = ((*s - 128) * 359) >> 8;
			cg = (cg + (*s++ - 128) * 183) >> 8;
			y2 = *s++;

			r = y1 + cr;
			g = y1 - cg;
			b = y1 + cb;
			SAT(r);
			SAT(g);
			SAT(b);

			*d++ = r;
			*d++ = g;
			*d++ = b;

			r = y2 + cr;
			g = y2 - cg;
			b = y2 + cb;
			SAT(r);
			SAT(g);
			SAT(b);

			*d++ = r;
			*d++ = g;
			*d++ = b;
		}
	}
}

// FOURCC "YUV9".
// To test and optimize...
static __inline__ void yuv410p_2_rgb24(int width, int height, unsigned char* src, unsigned char* dst)
{
	int r = 0, g = 0, b = 0, u = 0, y = 0, v = 0, u0 = 0, v0 = 0, cr = 0, cg = 0, cb = 0;
	unsigned char* buf = src;
	unsigned char* rgb = dst;
	int i = 0, j = 0, k = 0, n = 0, ki = 0, kj = 0, nbpix = width*height;

	// The Y plane is first. The Y plane has one byte per pixel. 
	// The Cb (U) plane immediately follows the Y plane in memory. 
	// The Cb plane is 1/4 the width ands 1/4 the height of the Y plane (and of the image). 
	// Following the Cb plane is the Cr (V) plane, just like the Cb plane. 
	for (i = 0; i <= height-4; i += 4) 
	{
		for (j = 0; j <= width-4; j += 4) 
		{
			u = buf[nbpix+n];
			v = buf[nbpix+nbpix/16+n];
			u0 = u-128; v0 = v-128; cr = (359*v0)>>8; cg = (88*u0+183*v0)>>8; cb = (454*u0)>>8;
			for (ki = i; ki < i+4; ki++) 
			{
				for (kj = j; kj < j+4; kj++) 
				{
					k = kj+width*ki;
					y = buf[k];

					r = y+cr; g = y-cg;	b = y+cb; SAT(r); SAT(g); SAT(b);

					rgb[3*k] = (unsigned char)r;
					rgb[3*k+1] = (unsigned char)g;
					rgb[3*k+2] = (unsigned char)b;
				}
			}
			n++;
		}
	}
}

// FOURCC "YVU9".
// To test and optimize...
static __inline__ void yvu410p_2_rgb24(int width, int height, unsigned char* src, unsigned char* dst)
{
	int r = 0, g = 0, b = 0, u = 0, y = 0, v = 0, u0 = 0, v0 = 0, cr = 0, cg = 0, cb = 0;
	unsigned char* buf = src;
	unsigned char* rgb = dst;
	int i = 0, j = 0, k = 0, n = 0, ki = 0, kj = 0, nbpix = width*height;

	for (i = 0; i <= height-4; i += 4) 
	{
		for (j = 0; j <= width-4; j += 4) 
		{
			v = buf[nbpix+n];
			u = buf[nbpix+nbpix/16+n];
			u0 = u-128; v0 = v-128; cr = (359*v0)>>8; cg = (88*u0+183*v0)>>8; cb = (454*u0)>>8;
			for (ki = i; ki < i+4; ki++) 
			{
				for (kj = j; kj < j+4; kj++) 
				{
					k = kj+width*ki;
					y = buf[k];

					r = y+cr; g = y-cg;	b = y+cb; SAT(r); SAT(g); SAT(b);

					rgb[3*k] = (unsigned char)r;
					rgb[3*k+1] = (unsigned char)g;
					rgb[3*k+2] = (unsigned char)b;
				}
			}
			n++;
		}
	}
}

// FOURCCs "YU12", "I420", "IYUV".
// For tests...
static __inline__ void i420_2_rgb24(int width, int height, unsigned char* src, unsigned char* dst)
{
	int r = 0, g = 0, b = 0, u = 0, y = 0, v = 0, u0 = 0, v0 = 0, cr = 0, cg = 0, cb = 0;
	unsigned char* buf = src;
	unsigned char* rgb = dst;
	int i = 0, j = 0, k = 0, n = 0, ki = 0, kj = 0, nbpix = width*height;

	// The Y plane is first. The Y plane has one byte per pixel. 
	// The Cb (U) plane immediately follows the Y plane in memory. 
	// The Cb plane is 1/2 the width ands 1/2 the height of the Y plane (and of the image). 
	// Following the Cb plane is the Cr (V) plane, just like the Cb plane. 
	for (i = 0; i <= height-2; i += 2) 
	{
		for (j = 0; j <= width-2; j += 2) 
		{
			u = buf[nbpix+n];
			v = buf[nbpix+nbpix/4+n];
			u0 = u-128; v0 = v-128; cr = (359*v0)>>8; cg = (88*u0+183*v0)>>8; cb = (454*u0)>>8;
			for (ki = i; ki < i+2; ki++) 
			{
				for (kj = j; kj < j+2; kj++) 
				{
					k = kj+width*ki;
					y = buf[k];

					r = y+cr; g = y-cg;	b = y+cb; SAT(r); SAT(g); SAT(b);

					rgb[3*k] = (unsigned char)r;
					rgb[3*k+1] = (unsigned char)g;
					rgb[3*k+2] = (unsigned char)b;
				}
			}
			n++;
		}
	}
}

// FOURCC "422P".
// To test and optimize...
static __inline__ void yuv422p_2_rgb24(int width, int height, unsigned char* src, unsigned char* dst)
{
	int r = 0, g = 0, b = 0, u = 0, y = 0, v = 0, u0 = 0, v0 = 0, cr = 0, cg = 0, cb = 0;
	unsigned char* buf = src;
	unsigned char* rgb = dst;
	int k = 0, n = 0, nbpix = width*height;

	// The Y plane is first. The Y plane has one byte per pixel.
	// The Cb (U) plane immediately follows the Y plane in memory. Each Cb belongs to two pixels.
	// Following the Cb plane is the Cr (V) plane, just like the Cb plane.
	for (k = 0; k <= nbpix-2; k += 2) 
	{
		y = buf[k];
		u = buf[nbpix+n];
		v = buf[nbpix+nbpix/2+n];

		u0 = u-128; v0 = v-128; cr = (359*v0)>>8; cg = (88*u0+183*v0)>>8; cb = (454*u0)>>8;
		r = y+cr; g = y-cg;	b = y+cb; SAT(r); SAT(g); SAT(b);

		rgb[3*k] = (unsigned char)r;
		rgb[3*k+1] = (unsigned char)g;
		rgb[3*k+2] = (unsigned char)b;

		y = buf[k+1];
		//u = buf[nbpix+n];
		//v = buf[nbpix+nbpix/2+n];

		//u0 = u-128; v0 = v-128; cr = (359*v0)>>8; cg = (88*u0+183*v0)>>8; cb = (454*u0)>>8;
		r = y+cr; g = y-cg;	b = y+cb; SAT(r); SAT(g); SAT(b);

		rgb[3*(k+1)] = (unsigned char)r;
		rgb[3*(k+1)+1] = (unsigned char)g;
		rgb[3*(k+1)+2] = (unsigned char)b;

		n++;
	}
}

// FOURCC "Y41P".
// To test and optimize...
static __inline__ void yuv411_2_rgb24(int width, int height, unsigned char* src, unsigned char* dst)
{
	int r = 0, g = 0, b = 0, u = 0, y = 0, v = 0, u0 = 0, v0 = 0, cr = 0, cg = 0, cb = 0;
	unsigned char* buf = src;
	unsigned char* rgb = dst;
	int k = 0, nbpix = width*height;

	for (k = 0; k <= nbpix-4; k += 4) 
	{
		u = buf[0];
		y = buf[1];
		v = buf[3];

		u0 = u-128; v0 = v-128; cr = (359*v0)>>8; cg = (88*u0+183*v0)>>8; cb = (454*u0)>>8;
		r = y+cr; g = y-cg;	b = y+cb; SAT(r); SAT(g); SAT(b);

		*rgb = (unsigned char)r; rgb++;
		*rgb = (unsigned char)g; rgb++;
		*rgb = (unsigned char)b; rgb++;

		//u = buf[0];
		y = buf[2];
		//v = buf[3];

		//u0 = u-128; v0 = v-128; cr = (359*v0)>>8; cg = (88*u0+183*v0)>>8; cb = (454*u0)>>8;
		r = y+cr; g = y-cg;	b = y+cb; SAT(r); SAT(g); SAT(b);

		*rgb = (unsigned char)r; rgb++;
		*rgb = (unsigned char)g; rgb++;
		*rgb = (unsigned char)b; rgb++;

		//u = buf[0];
		y = buf[4];
		//v = buf[3];

		//u0 = u-128; v0 = v-128; cr = (359*v0)>>8; cg = (88*u0+183*v0)>>8; cb = (454*u0)>>8;
		r = y+cr; g = y-cg;	b = y+cb; SAT(r); SAT(g); SAT(b);

		*rgb = (unsigned char)r; rgb++;
		*rgb = (unsigned char)g; rgb++;
		*rgb = (unsigned char)b; rgb++;

		//u = buf[0];
		y = buf[5];
		//v = buf[3];

		//u0 = u-128; v0 = v-128; cr = (359*v0)>>8; cg = (88*u0+183*v0)>>8; cb = (454*u0)>>8;
		r = y+cr; g = y-cg;	b = y+cb; SAT(r); SAT(g); SAT(b);

		*rgb = (unsigned char)r; rgb++;
		*rgb = (unsigned char)g; rgb++;
		*rgb = (unsigned char)b; rgb++;

		buf += 6;                                       
	}
}

// FOURCC "RGBO".
// To test and optimize...
static __inline__ void rgb555_2_rgb24(int width, int height, unsigned char* src, unsigned char* dst)
{
	int r = 0, g = 0, b = 0;
	unsigned char* buf = src;
	unsigned char* rgb = dst;
	int k = width*height;

	while (k)
	{
		k--;

		r = (buf[1]&0x7C)<<1;                         
		g = (((buf[0]&0xE0)>>2)|((buf[1]&0x03)<<6));
		b = (buf[0]&0x1F)<<3;               
		buf += 2;                                       

		// Pixel width*height-1-k.
		*rgb = (unsigned char)r; rgb++;
		*rgb = (unsigned char)g; rgb++;
		*rgb = (unsigned char)b; rgb++;
	}
}

// FOURCC "RGBP".
static __inline__ void rgb565_2_rgb24(int width, int height, unsigned char* src, unsigned char* dst)
{
	unsigned short tmp = 0;
	int r = 0, g = 0, b = 0;
	unsigned char* buf = src;
	unsigned char* rgb = dst;
	int k = width*height;

	while (k)
	{
		k--;

		tmp = *(unsigned short*)buf;    
		r = (tmp&0xF800)>>8;                               
		g = ((tmp<<5)&0xFC00)>>8;                          
		b = ((tmp<<11)&0xF800)>>8;                         
		buf += 2;                                       

		// Pixel width*height-1-k.
		*rgb = (unsigned char)r; rgb++;
		*rgb = (unsigned char)g; rgb++;
		*rgb = (unsigned char)b; rgb++;
	}
}

// FOURCC "RGB4".
static __inline__ void rgb32_2_rgb24(int width, int height, unsigned char* src, unsigned char* dst)
{
	unsigned char r = 0, g = 0, b = 0;
	unsigned char* buf = src;
	unsigned char* rgb = dst;
	int k = width*height;

	while (k)
	{
		k--;

		r = *buf; buf++;
		g = *buf; buf++;        
		b = *buf; buf++;    
		buf++;                                       

		// Pixel width*height-1-k.
		*rgb = r; rgb++;
		*rgb = g; rgb++;   
		*rgb = b; rgb++;                           
	}
}

// FOURCC "BGR4".
static __inline__ void bgr32_2_rgb24(int width, int height, unsigned char* src, unsigned char* dst)
{
	unsigned char r = 0, g = 0, b = 0;
	unsigned char* buf = src;
	unsigned char* rgb = dst;
	int k = width*height;

	while (k)
	{
		k--;

		b = *buf; buf++;
		g = *buf; buf++;        
		r = *buf; buf++;    
		buf++;                                       

		// Pixel width*height-1-k.
		*rgb = r; rgb++;
		*rgb = g; rgb++;   
		*rgb = b; rgb++;                           
	}
}

// FOURCC "RGB3".
static __inline__ void rgb24_2_rgb24(int width, int height, unsigned char* src, unsigned char* dst)
{
	unsigned char r = 0, g = 0, b = 0;
	unsigned char* buf = src;
	unsigned char* rgb = dst;
	int k = width*height;

	while (k)
	{
		k--;

		r = *buf; buf++;
		g = *buf; buf++;        
		b = *buf; buf++;    

		// Pixel width*height-1-k.
		*rgb = r; rgb++;
		*rgb = g; rgb++;   
		*rgb = b; rgb++;                           
	}
}

// FOURCC "BGR3".
static __inline__ void bgr24_2_rgb24(int width, int height, unsigned char* src, unsigned char* dst)
{
	unsigned char r = 0, g = 0, b = 0;
	unsigned char* buf = src;
	unsigned char* rgb = dst;
	int k = width*height;

	while (k)
	{
		k--;

		b = *buf; buf++;
		g = *buf; buf++;        
		r = *buf; buf++;    

		// Pixel width*height-1-k.
		*rgb = r; rgb++;
		*rgb = g; rgb++;   
		*rgb = b; rgb++;                           
	}
}

// FOURCCs "GREY", "Y800", "Y8  ".
// To test and optimize...
static __inline__ void grey8_2_rgb24(int width, int height, unsigned char* src, unsigned char* dst)
{
	unsigned char r = 0, g = 0, b = 0;
	unsigned char* buf = src;
	unsigned char* rgb = dst;
	int k = width*height;

	while (k)
	{
		k--;

		r = g = b = (*buf++);

		// Pixel width*height-1-k.
		*rgb = r; rgb++;
		*rgb = g; rgb++;   
		*rgb = b; rgb++;                           
	}
}

// FOURCC "Y16 ".
// To test and optimize...
static __inline__ void grey16_2_rgb24(int width, int height, unsigned char* src, unsigned char* dst)
{
	int r = 0, g = 0, b = 0;
	unsigned char* buf = src;
	unsigned char* rgb = dst;
	int k = width*height;

	while (k)
	{
		k--;

		r = g = b = ((*(unsigned short*)buf)>>8);      
		buf += 2;                       

		// Pixel width*height-1-k.
		*rgb = (unsigned char)r; rgb++;
		*rgb = (unsigned char)g; rgb++;
		*rgb = (unsigned char)b; rgb++;
	}
}

static __inline__ void SwapRedBluePPM(unsigned char* ppmimgdata, int width, int height)
{
	unsigned char c = 0;
	unsigned int i = 0;
	unsigned int size = width*height*3;

	while (i <= size-3)
	{
		c = ppmimgdata[i]; ppmimgdata[i] = ppmimgdata[i+2]; ppmimgdata[i+2] = c;
		i += 3;
	}
}

static __inline__ void ConvertV4LFormatToRGB24(
	unsigned char* webcamimgdata, unsigned char* ppmimgdata, 
	int width, int height, int depth, int palette, unsigned int minbuffersize)
{
	unsigned char* buf = webcamimgdata;
	unsigned char* rgb = ppmimgdata;

	switch (palette)
	{
	case VIDEO_PALETTE_GREY:
		switch (depth)
		{ 
		case 8:
			grey8_2_rgb24(width, height, buf, rgb); // Not tested...
			break;
		case 16:
			grey16_2_rgb24(width, height, buf, rgb); // Not tested...
			break;
		default:
			memcpy(rgb, buf, minbuffersize); // To try to get something... 
			break;
		}
		break;
	case VIDEO_PALETTE_HI240: 
		grey8_2_rgb24(width, height, buf, rgb); // Might get better than a black image...
		break;
	case VIDEO_PALETTE_RGB565:
		rgb565_2_rgb24(width, height, buf, rgb);
		break;
	case VIDEO_PALETTE_RGB24:
		bgr24_2_rgb24(width, height, buf, rgb);
		break;
	case VIDEO_PALETTE_RGB32:
		bgr32_2_rgb24(width, height, buf, rgb);
		break;
	case VIDEO_PALETTE_RGB555:
		rgb555_2_rgb24(width, height, buf, rgb); // Not tested...
		break;
	case VIDEO_PALETTE_YUV422: 
		//yuv422_2_rgb24(width, height, buf, rgb);
		yuyv_2_rgb24(width, height, buf, rgb); // YUV422 is the same as YUYV?
		break;
	case VIDEO_PALETTE_YUYV: 
		yuyv_2_rgb24(width, height, buf, rgb);
		break;
	case VIDEO_PALETTE_UYVY: 
		uyvy_2_rgb24(width, height, buf, rgb);
		break;
	case VIDEO_PALETTE_YUV420: 
		yuv420_2_rgb24(width, height, buf, rgb);
		break;
	case VIDEO_PALETTE_YUV411: 
		yuv411_2_rgb24(width, height, buf, rgb); // Not tested...
		break;
	case VIDEO_PALETTE_RAW: 
		memcpy(rgb, buf, minbuffersize); // To try to get something... 
		break;
	case VIDEO_PALETTE_YUV422P: 
		yuv422p_2_rgb24(width, height, buf, rgb); // Not tested...
		break;
	case VIDEO_PALETTE_YUV411P: 
		yuv411p_2_rgb24(width, height, buf, rgb); // Does not seem to be OK or maybe bad driver...
		break;
	case VIDEO_PALETTE_YUV420P: 
		yuv420p_2_rgb24(width, height, buf, rgb);
		//i420_2_rgb24(width, height, buf, rgb); // For tests...
		break;
	case VIDEO_PALETTE_YUV410P: 
		yuv410p_2_rgb24(width, height, buf, rgb); // Not tested...
		break;
	default:
		memcpy(rgb, buf, minbuffersize); // To try to get something... 
		break; 
	}
}

#endif // CONVERSIONS_H
