// C Standard Library headers.
#include <stdio.h> 
#include <stdlib.h> 
#include <string.h> 
#include <errno.h> 

// Linux headers.
#include <fcntl.h>
#include <malloc.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <linux/types.h>
#include <linux/videodev.h>

#include "Conversions.h"

// Globals.
int fd = 0;
struct video_capability vid_cap;
struct video_channel vid_chan;
struct video_window vid_win;
struct video_picture vid_pict;
struct video_mbuf vid_mb;
struct video_mmap* vid_mmaps;
void* mem = NULL;
int overlay = 0;

// PPM format is usually RGB 24 bits.
// minbuffersize should only be used in case the application cannot convert the data from the webcam.
void ConvertAndSaveToPPM(char* szFilename, unsigned char* webcamimgdata, 
	int width, int height, int depth, int palette, unsigned int minbuffersize)
{
	FILE* fp = NULL;
	unsigned char* ppmimgdata = NULL;

	fp = fopen(szFilename, "wb");
	if (fp)
	{
		ppmimgdata = (unsigned char*)calloc(width*height*3, sizeof(unsigned char));
		if (ppmimgdata)
		{
			fprintf(fp, "P6\n%d %d 255\n", width, height);
			ConvertV4LFormatToRGB24(webcamimgdata, ppmimgdata, width, height, depth, palette, minbuffersize);
			fwrite(ppmimgdata, sizeof(unsigned char), width*height*3, fp);
			free(ppmimgdata);
		}
		fclose(fp);
	}
	else
	{
		printf("fopen() warning : %s. \n", strerror(errno));
	}
}

void print_video_palette_name(int palette)
{
	switch (palette)
	{
	case VIDEO_PALETTE_GREY: printf("GREY"); break;
	case VIDEO_PALETTE_HI240: printf("HI240"); break;
	case VIDEO_PALETTE_RGB565: printf("RGB565"); break;
	case VIDEO_PALETTE_RGB24: printf("RGB24"); break;
	case VIDEO_PALETTE_RGB32: printf("RGB32"); break;
	case VIDEO_PALETTE_RGB555: printf("RGB555"); break;
	case VIDEO_PALETTE_YUV422: printf("YUV422"); break;
	case VIDEO_PALETTE_YUYV: printf("YUYV"); break;
	case VIDEO_PALETTE_UYVY: printf("UYVY"); break;
	case VIDEO_PALETTE_YUV420: printf("YUV420"); break;
	case VIDEO_PALETTE_YUV411: printf("YUV411"); break;
	case VIDEO_PALETTE_RAW: printf("RAW"); break;
	case VIDEO_PALETTE_YUV422P: printf("YUV422P"); break;
	case VIDEO_PALETTE_YUV411P: printf("YUV411P"); break;
	case VIDEO_PALETTE_YUV420P: printf("YUV420P"); break;
	case VIDEO_PALETTE_YUV410P: printf("YUV410P"); break;
	default: printf("Unknown"); break;
	}
}

void CleanUp(void)
{
	free(vid_mmaps);
	munmap(mem, vid_mb.size);
	overlay = 0;
	ioctl(fd, VIDIOCCAPTURE, &overlay);
	usleep(100000);
	close(fd);
}

void CleanUpWithoutMem(void)
{
	overlay = 0;
	ioctl(fd, VIDIOCCAPTURE, &overlay);
	usleep(100000);
	close(fd);
}

int main()
{
	int i = 0, j = 0, framequeue = 0;
	unsigned int count = 0;
	size_t size = 0;
	char szFilename[256];
	struct video_window vid_win_saved;
	struct video_picture vid_pict_saved;

	// Parameters.
	char szDevice[256];
	int bEnumCapabilities = 0;
	int bEnumChannels = 0;
	int bEnumSupportedChannelsNorms = 0;
	int bEnableChannelAndNormChoice = 0;
	int channel = 0;
	int norm = 0;
	int bEnableShowCaptureWindowSettings = 0;
	int bEnumSupportedResolutions = 0;
	int step = 0;
	int bEnableResolutionChoice = 0;
	int width = 0;
	int height = 0;
	int bEnableShowImageProperties = 0;
	int bEnumSupportedDepthAndPalette = 0;
	int bEnableDepthAndPaletteChoice = 0;
	int depth = 0;
	int palette = 0;
	int bEnableStandardControlsChanges = 0;
	int brightness = 0;
	int hue = 0;
	int colour = 0;
	int contrast = 0;
	int whiteness = 0;
	int bEnableOverlay = 0;
	int bUseOptimizedCaptureLoop = 0;
	unsigned int maxcount = 0;
	int bUseDeepestFrameQueue = 0;
	int nbframes = 0;
	int mbframesize = 0;
	int mbsize = 0;
	int bForceUserValues = 0;
	unsigned int minbuffersize = 0;

	// Options to be tweaked by user : 
	// _ szDevice = /dev/video0 ([IMPORTANT] from /dev/video0 to /dev/video63 for a webcam)
	// _ bEnumCapabilities = 1 (1 to enable, 0 to disable)
	// _ bEnumChannels = 1 (1 to enable, 0 to disable, bEnableChannelAndNormChoice should be set to 1 if enabled as the 
	// enumeration may change the current setting)
	// _ bEnumSupportedChannelsNorms = 1 (1 to enable, 0 to disable, bEnableChannelAndNormChoice should be set to 1 if 
	// enabled as the enumeration may change the current setting)
	// _ bEnableChannelAndNormChoice = 1 (1 to enable, 0 to disable)
	// _ channel = 1 ([IMPORTANT FOR TV CARDS] usually not used for a webcam, depends on the device if it has several inputs, 
	// starts at 0, valid only if bEnableChannelAndNormChoice is enabled)
	// _ norm = VIDEO_MODE_PAL (usually not used for a webcam, VIDEO_MODE_PAL (0), VIDEO_MODE_NTSC (1), VIDEO_MODE_SECAM (2) 
	//  or VIDEO_MODE_AUTO (3),you should tweak this setting if you have a TV card and the image has colors or synchronization 
	// problems, valid only if bEnableChannelAndNormChoice is enabled)
	// _ bEnableShowCaptureWindowSettings = 1 (1 to enable, 0 to disable)
	// _ bEnumSupportedResolutions = 0 (1 to enable, 0 to disable, valid only if bEnableShowCaptureWindowSettings is enabled)
	// _ step = 1 (width and height steps for the resolution enumeration, lower until you find a supported resolution if needed, 
	// or increase if the enumeration is too long)
	// _ bEnableResolutionChoice = 0 (1 to enable, 0 to disable)
	// _ width = 320 (one of the most common resolution is 320x240 pixels, might be 160x120, 320x240, 640x480, 800x600..., 
	// the driver should adjust it automatically to match device capabilities, valid only if bEnableResolutionChoice is enabled)
	// _ height = 240 (one of the most common resolution is 320x240 pixels, might be 160x120, 320x240, 640x480, 800x600..., 
	// the driver should adjust it automatically to match device capabilities, valid only if bEnableResolutionChoice is enabled)
	// _ bEnableShowImageProperties = 1 (1 to enable, 0 to disable)
	// _ bEnumSupportedDepthAndPalette = 1 (1 to enable, 0 to disable, valid only if bEnableShowImageProperties is enabled)
	// _ bEnableDepthAndPaletteChoice = 1 (1 to enable, 0 to disable)
	// _ depth = 16 ([IMPORTANT] number of bits per pixel, depends on the palette, valid values might be 4, 6, 8, 16, 24, 32..., 
	// valid only if bEnableDepthAndPaletteChoice is enabled)
	// _ palette = VIDEO_PALETTE_YUYV ([IMPORTANT] VIDEO_PALETTE_GREY (1), VIDEO_PALETTE_HI240 (2), VIDEO_PALETTE_RGB565 (3), 
	// VIDEO_PALETTE_RGB555 (4), VIDEO_PALETTE_RGB24 (5), VIDEO_PALETTE_RGB32 (6), VIDEO_PALETTE_YUV422 (7), VIDEO_PALETTE_YUYV (8), 
	// VIDEO_PALETTE_UYVY (9), VIDEO_PALETTE_YUV420 (10), VIDEO_PALETTE_YUV411 (11), VIDEO_PALETTE_RAW (12), VIDEO_PALETTE_YUV422P (13), 
	// VIDEO_PALETTE_YUV411P (14), VIDEO_PALETTE_YUV420P (15) or VIDEO_PALETTE_YUV410P (16), see the documentation of the corresponding 
	// four character code to know how to set the depth (maybe set the depth to 32 if you do not know), enable the enumeration of 
	// supported depth and palette to see a list of values, you should tweak this setting if you get black images, valid only 
	// if bEnableDepthAndPaletteChoice is enabled)
	// _ bEnableStandardControlsChanges = 0 (1 to enable, 0 to disable, valid only if bEnableDepthAndPaletteChoice is enabled)
	// _ brightness = 32768 (between 0 and 65535, valid only if bEnableStandardControlsChanges is enabled)
	// _ hue = 32768 (between 0 and 65535, valid only if bEnableStandardControlsChanges is enabled)
	// _ colour = 32768 (between 0 and 65535, valid only if bEnableStandardControlsChanges is enabled)
	// _ contrast = 32768 (between 0 and 65535, valid only if bEnableStandardControlsChanges is enabled)
	// _ whiteness = 32768 (between 0 and 65535, valid only if bEnableStandardControlsChanges is enabled)
	// _ bEnableOverlay = 0 (1 to enable, 0 to disable, might only be useful when direct display on the screen is needed)
	// _ bUseOptimizedCaptureLoop = 1 (1 to enable, 0 to disable, change if you do not get any image)
	// _ maxcount = 100 (number of frames to capture)
	// _ bUseDeepestFrameQueue = 0 (1 to enable, 0 to disable, valid only if bUseOptimizedCaptureLoop is enabled)
	// _ nbframes = 1 (in case the driver does not indicate the right number of frames in the memory map)
	// _ mbframesize = width*height*((depth+7)/8) (in case the driver does not indicate the right frame size in the memory map)
	// _ mbsize = nbframes*mbframesize (in case the driver does not indicate the right total size in the memory map)
	// _ bForceUserValues = 0 (1 to enable, 0 to disable, might be useful if the webcam driver is not responding correctly)
	// _ minbuffersize = 0 (used if the application cannot convert the data from the camera and if different from 0, otherwise 
	// a default value is tried)
	sprintf(szDevice, "/dev/video0");
	bEnumCapabilities = 1;
	bEnumChannels = 1;
	bEnumSupportedChannelsNorms = 1;
	bEnableChannelAndNormChoice = 1;
	channel = 1;
	norm = VIDEO_MODE_PAL;
	bEnableShowCaptureWindowSettings = 1;
	bEnumSupportedResolutions = 0;
	step = 1;
	bEnableResolutionChoice = 0;
	width = 320;
	height = 240;
	bEnableShowImageProperties = 1;
	bEnumSupportedDepthAndPalette = 1;
	bEnableDepthAndPaletteChoice = 1;
	depth = 16;
	palette = VIDEO_PALETTE_YUYV;
	bEnableStandardControlsChanges = 0;
	brightness = 32768;
	hue = 32768;
	colour = 32768;
	contrast = 32768;
	whiteness = 32768;
	bEnableOverlay = 0;
	bUseOptimizedCaptureLoop = 1;
	maxcount = 100;
	bUseDeepestFrameQueue = 0;
	nbframes = 1;
	mbframesize = width*height*((depth+7)/8);
	mbsize = nbframes*mbframesize;
	bForceUserValues = 0;
	minbuffersize = 0;

	// Open the desired device.
	fd = open(szDevice, O_RDWR); // O_RDWR, O_RDONLY, O_EXCL, O_NONBLOCK...?
	if (fd == -1)
	{
		printf("open() error : %s. \n", strerror(errno));
		return EXIT_FAILURE;
	}	
	usleep(100000);

	if (bEnumCapabilities)
	{
		// Obtain the capability information for a video device.
		memset(&vid_cap, 0, sizeof(vid_cap));
		if (ioctl(fd, VIDIOCGCAP, &vid_cap) == -1)
		{
			printf("ioctl() VIDIOCGCAP warning : %s. \n", strerror(errno));
		}
		else
		{
			printf("Video device name : %s, number of channels : %d, number of audio devices : %d", vid_cap.name, vid_cap.channels, vid_cap.audios);
			if (vid_cap.type & VID_TYPE_CAPTURE) printf(", can capture to memory");
			if (vid_cap.type & VID_TYPE_TUNER) printf(", has a tuner");
			if (vid_cap.type & VID_TYPE_TELETEXT) printf(", has teletext");
			if (vid_cap.type & VID_TYPE_OVERLAY) printf(", can overlay its image onto the frame buffer");
			if (vid_cap.type & VID_TYPE_CHROMAKEY) printf(", overlay is Chromakeyed");
			if (vid_cap.type & VID_TYPE_CLIPPING) printf(", overlay clipping is supported");
			if (vid_cap.type & VID_TYPE_FRAMERAM) printf(", overlay overwrites frame buffer memory");
			if (vid_cap.type & VID_TYPE_SCALES) printf(", supports image scaling");
			if (vid_cap.type & VID_TYPE_MONOCHROME) printf(", is grey scale only");
			if (vid_cap.type & VID_TYPE_SUBCAPTURE) printf(", capture can be of only part of the image");
			printf(". \n");
		}
		usleep(100000);
	}

	// There seem to be some differences between the documentation and the implementation around VIDIOCSCHAN/VIDIOCGCHAN...

	if (bEnumChannels)
	{
		// Get info about the different channels of the device.
		for (i = 0; i < vid_cap.channels; i++)
		{
			// Select channel for which we want to get info.
			memset(&vid_chan, 0, sizeof(vid_chan));
			vid_chan.channel = i;
			if (ioctl(fd, VIDIOCGCHAN, &vid_chan) == -1)
			{
				printf("ioctl() VIDIOCGCHAN warning : %s. \n", strerror(errno));
				continue;
			}
			usleep(100000);
			printf("Video device channel number : %d, name : %s, number of tuners : %d, norm identifier : %d", 
				vid_chan.channel, vid_chan.name, vid_chan.tuners, (int)vid_chan.norm);
			if (vid_chan.flags & VIDEO_VC_TUNER) printf(", has tuners");
			if (vid_chan.flags & VIDEO_VC_AUDIO) printf(", has audio");
			//if (vid_chan.flags & VIDEO_VC_NORM) printf(", has norm setting");
			if (vid_chan.type & VIDEO_TYPE_TV) printf(", is a TV input");
			if (vid_chan.type & VIDEO_TYPE_CAMERA) printf(", is a camera");
			if (vid_chan.norm == VIDEO_MODE_PAL) printf(", norm : PAL");
			if (vid_chan.norm == VIDEO_MODE_NTSC) printf(", norm : NTSC");
			if (vid_chan.norm == VIDEO_MODE_SECAM) printf(", norm : SECAM");
			if (vid_chan.norm == VIDEO_MODE_AUTO) printf(", norm : AUTO");
			if (bEnumSupportedChannelsNorms)
			{
				printf(", supported norms : \n");
				vid_chan.norm = VIDEO_MODE_PAL; if ((ioctl(fd, VIDIOCSCHAN, &vid_chan) != -1)&&(ioctl(fd, VIDIOCGCHAN, &vid_chan) != -1)&&((int)vid_chan.norm == VIDEO_MODE_PAL)) printf("     PAL\n"); usleep(100000);
				vid_chan.norm = VIDEO_MODE_NTSC; if ((ioctl(fd, VIDIOCSCHAN, &vid_chan) != -1)&&(ioctl(fd, VIDIOCGCHAN, &vid_chan) != -1)&&((int)vid_chan.norm == VIDEO_MODE_NTSC)) printf("     NTSC\n"); usleep(100000);
				vid_chan.norm = VIDEO_MODE_SECAM; if ((ioctl(fd, VIDIOCSCHAN, &vid_chan) != -1)&&(ioctl(fd, VIDIOCGCHAN, &vid_chan) != -1)&&((int)vid_chan.norm == VIDEO_MODE_SECAM)) printf("     SECAM\n"); usleep(100000);
				vid_chan.norm = VIDEO_MODE_AUTO; if ((ioctl(fd, VIDIOCSCHAN, &vid_chan) != -1)&&(ioctl(fd, VIDIOCGCHAN, &vid_chan) != -1)&&((int)vid_chan.norm == VIDEO_MODE_AUTO)) printf("     AUTO\n"); usleep(100000);
			}
			else
			{
				printf(". \n");
			}
		}
		printf("\n");
	}

	if (bEnableChannelAndNormChoice)
	{
		// Choose a channel (usually channel 0 for a webcam).
		// Need also to choose a mode/norm (VIDEO_MODE_PAL, VIDEO_MODE_NTSC, VIDEO_MODE_SECAM or VIDEO_MODE_AUTO)?
		// First get channel info and then select channel and change options?
		memset(&vid_chan, 0, sizeof(vid_chan));
		vid_chan.channel = channel;
		if (ioctl(fd, VIDIOCGCHAN, &vid_chan) == -1)
		{
			printf("ioctl() VIDIOCGCHAN warning : %s. \n", strerror(errno));
			memset(&vid_chan, 0, sizeof(vid_chan));
		}
		usleep(100000);
		vid_chan.channel = channel;
		vid_chan.norm = (__u16)norm;
		if (ioctl(fd, VIDIOCSCHAN, &vid_chan) == -1)
		{
			printf("ioctl() VIDIOCSCHAN warning : %s. \n", strerror(errno));
		}
		usleep(100000);
		// Check if the settings requested were accepted (actually, we can just check the norm...).
		memset(&vid_chan, 0, sizeof(vid_chan));
		vid_chan.channel = channel;
		if (ioctl(fd, VIDIOCGCHAN, &vid_chan) == -1)
		{
			printf("ioctl() VIDIOCGCHAN warning : %s. \n", strerror(errno));

			// Values that will be used in what follows.
			vid_chan.channel = channel;
			vid_chan.norm = (__u16)norm;
		}
		else
		{
			if ((vid_chan.channel != channel)||((int)vid_chan.norm != norm))
			{
				printf("An unsupported channel setting was specified. \n");
			}
		}
		usleep(100000);
		printf("Channel used : %d, norm identifier : %d", vid_chan.channel, (int)vid_chan.norm);
		if (vid_chan.norm == VIDEO_MODE_PAL) printf(", norm : PAL");
		if (vid_chan.norm == VIDEO_MODE_NTSC) printf(", norm : NTSC");
		if (vid_chan.norm == VIDEO_MODE_SECAM) printf(", norm : SECAM");
		if (vid_chan.norm == VIDEO_MODE_AUTO) printf(", norm : AUTO");
		printf(". \n");
	}

	if (bEnableShowCaptureWindowSettings)
	{
		// Capture window settings.
		memset(&vid_win, 0, sizeof(vid_win));
		if (ioctl(fd, VIDIOCGWIN, &vid_win) == -1)
		{
			printf("ioctl() VIDIOCGWIN warning : %s. \n", strerror(errno));
		}
		else
		{
			usleep(100000);
			printf("Video device current capture window x coordinate : %u, y coordinate : %u, width : %u, height : %u, "
				"host order RGB32 value for the chroma key : %u, flags : %u, "
				"pointer address to a list of clipping rectangles : %#x, number of clipping rectangles : %d. \n", 
				vid_win.x, vid_win.y, vid_win.width, vid_win.height, vid_win.chromakey, vid_win.flags, (unsigned int)vid_win.clips, vid_win.clipcount);

			if (bEnumSupportedResolutions)
			{
				// Save the current settings because the enumeration might change them.
				memcpy(&vid_win_saved, &vid_win, sizeof(vid_win));
				// Enumerate all supported resolutions.
				printf("Video device supported resolutions : \n");
				memset(&vid_win, 0, sizeof(vid_win));
				vid_win.x = 0;
				vid_win.y = 0;
				vid_win.chromakey = 0;
				vid_win.flags = 0; // ???
				vid_win.clips = NULL;
				vid_win.clipcount = 0;
				for (i = vid_cap.minheight; i <= vid_cap.maxheight; i += step)
				{
					for (j = vid_cap.minwidth; j <= vid_cap.maxwidth; j += step)
					{
						vid_win.height = (__u32)i;
						vid_win.width = (__u32)j;
						if ((ioctl(fd, VIDIOCSWIN, &vid_win) != -1)&&(ioctl(fd, VIDIOCGWIN, &vid_win) != -1)&&
							((int)vid_win.width == j)&&((int)vid_win.height == i))
						{
							printf("     %ux%u\n", vid_win.width, vid_win.height);
						}
					}
				}
				printf("\n");
				// Try to restore the previous settings.
				memcpy(&vid_win, &vid_win_saved, sizeof(vid_win));
				ioctl(fd, VIDIOCSWIN, &vid_win);
			}
		}
		usleep(100000);
	}

	if (bEnableResolutionChoice)
	{
		// Set the desired resolution.
		memset(&vid_win, 0, sizeof(vid_win));
		vid_win.x = 0;
		vid_win.y = 0;
		vid_win.chromakey = 0;
		vid_win.flags = 0; // ???
		vid_win.clips = NULL;
		vid_win.clipcount = 0;
		vid_win.width = (__u32)width;
		vid_win.height = (__u32)height;
		if (ioctl(fd, VIDIOCSWIN, &vid_win) == -1)
		{
			printf("ioctl() VIDIOCSWIN warning : %s. \n", strerror(errno));
		}
		usleep(100000);
		// Check if the resolution requested was accepted.
		memset(&vid_win, 0, sizeof(vid_win));
		if (ioctl(fd, VIDIOCGWIN, &vid_win) == -1)
		{
			printf("ioctl() VIDIOCGWIN warning : %s. \n", strerror(errno));

			// Values that will be used in what follows.
			vid_win.width = (__u32)width;
			vid_win.height = (__u32)height;
		}
		else
		{
			if (((int)vid_win.width != width)||((int)vid_win.height != height))
			{
				printf("An unsupported resolution was specified. \n");
			}
		}
		usleep(100000);
		printf("Resolution used : %ux%u. \n", vid_win.width, vid_win.height);
	}

	if (bEnableShowImageProperties)
	{
		// Image properties.
		memset(&vid_pict, 0, sizeof(vid_pict));
		if (ioctl(fd, VIDIOCGPICT, &vid_pict) == -1)
		{
			printf("ioctl() VIDIOCGPICT warning : %s. \n", strerror(errno));
		}
		else
		{
			usleep(100000);
			printf("Video device current image brightness : %d, hue : %d, colour : %d, contrast : %d, whiteness : %d, depth : %d, palette identifier : %d, palette : ", 
				(int)vid_pict.brightness, (int)vid_pict.hue, (int)vid_pict.colour, (int)vid_pict.contrast, (int)vid_pict.whiteness, (int)vid_pict.depth, (int)vid_pict.palette);
			print_video_palette_name((int)vid_pict.palette);
			printf(". \n");

			if (bEnumSupportedDepthAndPalette)
			{
				// Save the current settings because the enumeration might change them.
				memcpy(&vid_pict_saved, &vid_pict, sizeof(vid_pict));
				// Enumerate all supported depth and palette.
				// Only depths up to 32 bits per pixel are tested (which should be more than the maximum for most of the webcams).
				printf("Video device supported depth and palette : \n");
				for (i = 1; i <= 16; i++)
				{
					for (j = 1; j <= 32; j++)
					{
						vid_pict.palette = (__u16)i;
						vid_pict.depth = (__u16)j; 
						if ((ioctl(fd, VIDIOCSPICT, &vid_pict) != -1)&&(ioctl(fd, VIDIOCGPICT, &vid_pict) != -1)&&
							((int)vid_pict.depth == j)&&((int)vid_pict.palette == i))
						{
							printf("     %d bpp, ", (int)vid_pict.depth);
							print_video_palette_name((int)vid_pict.palette);
							printf("\n");
						}
					}
				}
				printf("\n");
				// Try to restore the previous settings.
				memcpy(&vid_pict, &vid_pict_saved, sizeof(vid_pict));
				ioctl(fd, VIDIOCSPICT, &vid_pict);
			}
		}
		usleep(100000);
	}

	if (bEnableDepthAndPaletteChoice)
	{
		// Get current image properties.
		memset(&vid_pict, 0, sizeof(vid_pict));
		if (ioctl(fd, VIDIOCGPICT, &vid_pict) == -1)
		{
			printf("ioctl() VIDIOCGPICT warning : %s. \n", strerror(errno));

			memset(&vid_pict, 0, sizeof(vid_pict));
		}
		// Set the desired image properties.
		if (bEnableStandardControlsChanges)
		{
			vid_pict.brightness = (__u16)brightness;
			vid_pict.hue = (__u16)hue;
			vid_pict.colour = (__u16)colour;
			vid_pict.contrast = (__u16)contrast;
			vid_pict.whiteness = (__u16)whiteness;
		}
		vid_pict.depth = (__u16)depth;
		vid_pict.palette = (__u16)palette;
		if (ioctl(fd, VIDIOCSPICT, &vid_pict) == -1)
		{
			printf("ioctl() VIDIOCSPICT warning : %s. \n", strerror(errno));
		}
		usleep(100000);
		// Check if the image properties requested were accepted.
		memset(&vid_pict, 0, sizeof(vid_pict));
		if (ioctl(fd, VIDIOCGPICT, &vid_pict) == -1)
		{
			printf("ioctl() VIDIOCGPICT warning : %s. \n", strerror(errno));

			// Values that will be used in what follows.
			if (bEnableStandardControlsChanges)
			{
				vid_pict.brightness = (__u16)brightness;
				vid_pict.hue = (__u16)hue;
				vid_pict.colour = (__u16)colour;
				vid_pict.contrast = (__u16)contrast;
				vid_pict.whiteness = (__u16)whiteness;
			}
			vid_pict.depth = (__u16)depth;
			vid_pict.palette = (__u16)palette;
		}
		else
		{
			if (
				(
				(bEnableStandardControlsChanges)&&(
				((int)vid_pict.brightness != brightness)||((int)vid_pict.hue != hue)||((int)vid_pict.colour != colour)||
				((int)vid_pict.contrast != contrast)||((int)vid_pict.whiteness != whiteness))
				)
				||((int)vid_pict.depth != depth)||((int)vid_pict.palette != palette)
				)
			{
				printf("An unsupported image property was specified. \n");
			}
		}
		usleep(100000);
		printf("Image brightness used : %d, hue : %d, colour : %d, contrast : %d, whiteness : %d, depth : %d, palette identifier : %d, palette : ", 
			(int)vid_pict.brightness, (int)vid_pict.hue, (int)vid_pict.colour, (int)vid_pict.contrast, (int)vid_pict.whiteness, (int)vid_pict.depth, (int)vid_pict.palette);
		print_video_palette_name((int)vid_pict.palette);
		printf(". \n");
	}

	if (bEnableOverlay)
	{
		// Enable overlay capturing.
		overlay = 1;
		if (ioctl(fd, VIDIOCCAPTURE, &overlay) == -1)
		{
			printf("ioctl() VIDIOCCAPTURE warning : %s. \n", strerror(errno));
		}
		usleep(100000);
	}

	if (bForceUserValues)
	{
		vid_chan.channel = channel;
		vid_chan.norm = (__u16)norm;
		vid_win.width = (__u32)width;
		vid_win.height = (__u32)height;
		vid_pict.depth = (__u16)depth;
		vid_pict.palette = (__u16)palette;
		printf("Because you selected to force user parameters instead of driver values, the following values will be used \n"
			"channel number : %d, norm identifier : %d, width : %u, height : %u, depth : %d, palette identifier : %d, palette : ", 
			vid_chan.channel, (int)vid_chan.norm, vid_win.width, vid_win.height, (int)vid_pict.depth, (int)vid_pict.palette);
		print_video_palette_name((int)vid_pict.palette);
		printf(". \n\n");
	}

	if (!bUseOptimizedCaptureLoop)
	{
		// Try the simple method with the read() function.
		size = vid_win.width*vid_win.height*(unsigned int)((vid_pict.depth+7)>>3);
		mem = calloc(size, sizeof(unsigned char));
		if (mem == NULL)
		{
			printf("calloc() error : %s. \n", strerror(errno));
			CleanUpWithoutMem();
			return EXIT_FAILURE;
		}

		// Simple capture loop.
		count = 0;
		while (count < maxcount)
		{
			// Maybe we should try to continue to fill the buffer with other calls to read() 
			// if the returned value is lower than size but we must be sure of the size...
			if (read(fd, mem, size) == -1)
			{
				printf("read() error : %s. \n", strerror(errno));
				free(mem);
				CleanUpWithoutMem();
				return EXIT_FAILURE;
			}
			sprintf(szFilename, "%.8d.ppm", count);
			ConvertAndSaveToPPM(szFilename, (unsigned char*)mem, (int)vid_win.width, 
				(int)vid_win.height, (int)vid_pict.depth, (int)vid_pict.palette, 
				minbuffersize?minbuffersize:size/256);
			count++;
		}

		free(mem);
	}
	else
	{
		size = vid_win.width*vid_win.height*(unsigned int)((vid_pict.depth+7)>>3);

		// Get memory buffer information.
		memset(&vid_mb, 0, sizeof(vid_mb));
		if (ioctl(fd, VIDIOCGMBUF, &vid_mb) == -1)
		{
			printf("ioctl() VIDIOCGMBUF warning : %s. \n", strerror(errno));

			// Values that will be used in what follows.
			vid_mb.size = mbsize;
			vid_mb.frames = nbframes;
			for (i = 0; i < vid_mb.frames; i++)
			{
				vid_mb.offsets[i] = i*mbframesize;
			}
		}

		if (bForceUserValues)
		{
			printf("The video device memory buffer values will be forced user parameters. \n");
			vid_mb.size = mbsize;
			vid_mb.frames = nbframes;
			for (i = 0; i < vid_mb.frames; i++)
			{
				vid_mb.offsets[i] = i*mbframesize;
			}
		}

		printf("Video device memory buffer number of bytes to map : %d, number of frames : %d, offset of each frame : \n", 
			vid_mb.size, vid_mb.frames);
		for (i = 0; i < vid_mb.frames; i++)
		{
			printf("     Frame %d : %d\n", i, vid_mb.offsets[i]);
		}
		printf("\n");

		// Set a memory mapping between driver memory and this program memory, where the image data should be available.
		// PROT_READ specifies that the memory mapped will only be read from, MAP_SHARED specifies that a 
		// change to the video memory will change the user space memory.
		mem = mmap(0, vid_mb.size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0); // PROT_READ, PROT_WRITE, PROT_EXEC, MAP_SHARED, MAP_EXECUTABLE...?
		if (mem == MAP_FAILED)
		{
			printf("mmap() warning : %s. \n", strerror(errno));

			// Will retry after an ioctl() VIDIOCMCAPTURE...
		}

		// Set information for each frame in the mmap in case the device can buffer several frames.
		// Maybe it is not necessary to allocate several struct video_mmap...
		vid_mmaps = (struct video_mmap*)calloc(vid_mb.frames, sizeof(struct video_mmap));
		if (vid_mmaps == NULL)
		{
			printf("calloc() error : %s. \n", strerror(errno));
			if (mem != MAP_FAILED) munmap(mem, vid_mb.size);
			CleanUpWithoutMem();
			return EXIT_FAILURE;
		}

		// Optimized capture loop.
		count = 0;
		framequeue = bUseDeepestFrameQueue?vid_mb.frames:1;
		for (i = 0; i < framequeue; i++)
		{
			memset(&vid_mmaps[i], 0, sizeof(vid_mmaps[i]));
			vid_mmaps[i].frame = (unsigned int)i;
			vid_mmaps[i].width = (int)vid_win.width;
			vid_mmaps[i].height = (int)vid_win.height;
			vid_mmaps[i].format = vid_pict.palette;
			if (ioctl(fd, VIDIOCMCAPTURE, &vid_mmaps[i]) == -1)
			{
				printf("ioctl() VIDIOCMCAPTURE warning : %s. \n", strerror(errno));
			}
		}
		while (count < maxcount)
		{
			for (i = 0; i < framequeue; i++) 
			{
				if (ioctl(fd, VIDIOCSYNC, &i) == -1)
				{
					printf("ioctl() VIDIOCSYNC warning : %s. \n", strerror(errno));
				}
				// Retry the mmap() in case it failed before...
				if (mem == MAP_FAILED) 
				{
					mem = mmap(0, vid_mb.size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
					if (mem == MAP_FAILED)
					{
						printf("mmap() error : %s. \n", strerror(errno));
						free(vid_mmaps);
						CleanUpWithoutMem();
						return EXIT_FAILURE;
					}
				}
				// Image data are available at mem+vid_mb.offsets[i].
				sprintf(szFilename, "%.8d.ppm", count);
				ConvertAndSaveToPPM(szFilename, (unsigned char*)mem+vid_mb.offsets[i], 
					(int)vid_win.width, (int)vid_win.height, (int)vid_pict.depth, (int)vid_pict.palette, 
					minbuffersize?minbuffersize:size/256);
				count++;
				// Next frame.
				memset(&vid_mmaps[i], 0, sizeof(vid_mmaps[i]));
				vid_mmaps[i].frame = (unsigned int)i;
				vid_mmaps[i].width = (int)vid_win.width;
				vid_mmaps[i].height = (int)vid_win.height;
				vid_mmaps[i].format = vid_pict.palette;
				if (ioctl(fd, VIDIOCMCAPTURE, &vid_mmaps[i]) == -1)
				{
					printf("ioctl() VIDIOCMCAPTURE warning : %s. \n", strerror(errno));
				}
			}
		}
		for (i = 0; i < framequeue; i++)
		{
			if (ioctl(fd, VIDIOCSYNC, &i) == -1)
			{
				printf("ioctl() VIDIOCSYNC warning : %s. \n", strerror(errno));
			}
		}

		free(vid_mmaps);

		if (munmap(mem, vid_mb.size) != EXIT_SUCCESS)
		{
			printf("munmap() error : %s. \n", strerror(errno));
			CleanUpWithoutMem();
			return EXIT_FAILURE;
		}
	}

	if (bEnableOverlay)
	{
		// Disable overlay capturing.
		overlay = 0;
		if (ioctl(fd, VIDIOCCAPTURE, &overlay) == -1)
		{
			printf("ioctl() VIDIOCCAPTURE warning : %s. \n", strerror(errno));
		}
		usleep(100000);
	}

	// Close the device.
	if (close(fd) != EXIT_SUCCESS)
	{
		printf("close() error : %s. \n", strerror(errno));
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
